@@ -17,9 +17,13 @@ use hyper::{Client as HttpClient, StatusCode, Uri};
 use log::{debug, error, warn};
 use std::net::SocketAddr;
 use tokio::{
+    join,
     net::TcpStream,
     select,
-    sync::mpsc::{channel, Receiver, Sender},
+    sync::{
+        mpsc::{channel, Receiver, Sender},
+        oneshot,
+    },
 };
 use tokio_tungstenite::{
     connect_async, tungstenite::Message as WsMessage, MaybeTlsStream, WebSocketStream,
@@ -30,12 +34,14 @@ use uuid::Uuid;
 mod channels {
     pub const EVENTS: usize = 100;
     pub const MESSAGES: usize = 100;
+    pub const SHUTDOWN: usize = 1;
 }
 
 #[derive(Debug, Clone)]
 pub struct Control {
     shutdown: CancellationToken,
     tx_sender: Sender<Message>,
+    tx_shutdown: Sender<oneshot::Sender<()>>,
 }
 
 impl Control {
@@ -47,8 +53,15 @@ impl Control {
 #[async_trait]
 impl client::Control<Error> for Control {
     async fn shutdown(&self) -> Result<(), Error> {
+        let (tx_response, rx_response): (oneshot::Sender<()>, oneshot::Receiver<()>) =
+            oneshot::channel();
+        self.tx_shutdown
+            .send(tx_response)
+            .await
+            .map_err(|e| Error::Channel(e.to_string()))?;
+        let result = rx_response.await;
         self.shutdown.cancel();
-        Ok(())
+        result.map_err(|e| Error::Channel(e.to_string()))
     }
     async fn send(&self, msg: Message) -> Result<(), Error> {
         self.tx_sender
@@ -63,6 +76,7 @@ pub struct Client {
     uuid: Uuid,
     control: Control,
     rx_sender: Option<Receiver<Message>>,
+    rx_shutdown: Option<Receiver<oneshot::Sender<()>>>,
     rx_events: Option<Receiver<Event<Error>>>,
     tx_events: Sender<Event<Error>>,
 }
@@ -71,6 +85,10 @@ impl Client {
     pub fn new(options: Options) -> Self {
         let (tx_sender, rx_sender): (Sender<Message>, Receiver<Message>) =
             channel(channels::MESSAGES);
+        let (tx_shutdown, rx_shutdown): (
+            Sender<oneshot::Sender<()>>,
+            Receiver<oneshot::Sender<()>>,
+        ) = channel(channels::SHUTDOWN);
         let (tx_events, rx_events): (Sender<Event<Error>>, Receiver<Event<Error>>) =
             channel(channels::EVENTS);
         Self {
@@ -78,9 +96,11 @@ impl Client {
             tx_events,
             rx_sender: Some(rx_sender),
             rx_events: Some(rx_events),
+            rx_shutdown: Some(rx_shutdown),
             uuid: Uuid::new_v4(),
             control: Control {
                 tx_sender,
+                tx_shutdown,
                 shutdown: CancellationToken::new(),
             },
         }
@@ -174,9 +194,12 @@ impl Client {
         mut reader: SplitStream<WebSocketStream<MaybeTlsStream<tokio::net::TcpStream>>>,
         tx_events: Sender<Event<Error>>,
         cancel: CancellationToken,
-    ) -> Result<(), Error> {
+    ) -> (
+        SplitStream<WebSocketStream<MaybeTlsStream<tokio::net::TcpStream>>>,
+        Option<Error>,
+    ) {
         debug!(target: logs::targets::CLIENT, "reader_task is started");
-        let result = select! {
+        let error = select! {
             res = async {
                 while let Some(msg) = reader.next().await {
                     match msg {
@@ -217,19 +240,38 @@ impl Client {
                     }
                 }
                 Ok(())
-            } => res,
-            _ = cancel.cancelled() => Ok(())
+            } => {
+                if let Err(error) = res {
+                    Some(error)
+                } else {
+                    None
+                }
+            },
+            _ = cancel.cancelled() => None
         };
+        if let Some(err) = error.clone() {
+            if let Err(err) = tx_events.send(Event::Error(err)).await {
+                error!(
+                    target: logs::targets::CLIENT,
+                    "fail to send event Error; error: {:?}", err
+                );
+            }
+        }
         debug!(target: logs::targets::CLIENT, "reader_task is finished");
-        result
+        cancel.cancel();
+        (reader, error)
     }
 
     async fn writer_task(
         mut writer: SplitSink<WebSocketStream<MaybeTlsStream<TcpStream>>, WsMessage>,
+        tx_events: Sender<Event<Error>>,
         mut rx_sender: Receiver<Message>,
         cancel: CancellationToken,
-    ) -> Result<(), Error> {
-        let result = select! {
+    ) -> (
+        SplitSink<WebSocketStream<MaybeTlsStream<TcpStream>>, WsMessage>,
+        Option<Error>,
+    ) {
+        let error = select! {
             res = async {
                 while let Some(msg) = rx_sender.recv().await {
                     match msg {
@@ -252,28 +294,43 @@ impl Client {
                     }
                 }
                 Ok(())
-            } => res,
-            _ = cancel.cancelled() => Ok(())
+            } => {
+                if let Err(error) = res {
+                    Some(error)
+                } else {
+                    None
+                }
+            },
+            _ = cancel.cancelled() => None
         };
-        if let Err(err) = writer.close().await {
-            error!(
-                target: logs::targets::CLIENT,
-                "fail to close socket; error: {}", err
-            );
+        if let Some(err) = error.clone() {
+            if let Err(err) = tx_events.send(Event::Error(err)).await {
+                error!(
+                    target: logs::targets::CLIENT,
+                    "fail to send event Error; error: {:?}", err
+                );
+            }
         }
-        result
+        cancel.cancel();
+        (writer, error)
     }
 
     fn reinit(&mut self) {
         let (tx_sender, rx_sender): (Sender<Message>, Receiver<Message>) =
             channel(channels::MESSAGES);
+        let (tx_shutdown, rx_shutdown): (
+            Sender<oneshot::Sender<()>>,
+            Receiver<oneshot::Sender<()>>,
+        ) = channel(channels::SHUTDOWN);
         let (tx_events, rx_events): (Sender<Event<Error>>, Receiver<Event<Error>>) =
             channel(channels::EVENTS);
         self.tx_events = tx_events;
         self.rx_events = Some(rx_events);
         self.rx_sender = Some(rx_sender);
+        self.rx_shutdown = Some(rx_shutdown);
         self.control = Control {
             tx_sender,
+            tx_shutdown,
             shutdown: CancellationToken::new(),
         };
         debug!(target: logs::targets::CLIENT, "client has been reinited");
@@ -285,7 +342,6 @@ impl Impl<Error, Control> for Client {
     async fn connect(&mut self) -> Result<(), Error> {
         env::logs::init();
         debug!(target: logs::targets::CLIENT, "client is started");
-        let cancel = self.control.get_shutdown_token();
         let socket = match self.options.connection {
             ConnectionType::Direct(addr) => {
                 Self::direct_connection(addr, self.tx_events.clone()).await
@@ -315,22 +371,67 @@ impl Impl<Error, Control> for Client {
         let rx_sender = if let Some(rx_sender) = self.rx_sender.take() {
             rx_sender
         } else {
-            return Err(Error::SenderAlreadyTaken);
+            return Err(Error::AlreadyInited);
         };
-        let res = select! {
-            res = Self::reader_task(reader, self.tx_events.clone(), cancel.child_token()) => res,
-            res = Self::writer_task(writer, rx_sender, cancel) => res,
+        let mut rx_shutdown = if let Some(rx_shutdown) = self.rx_shutdown.take() {
+            rx_shutdown
+        } else {
+            return Err(Error::AlreadyInited);
         };
+        let cancel = CancellationToken::new();
+        let (((reader, reader_error), (writer, writer_error)), tx_shutdown_response) = join!(
+            async {
+                join!(
+                    Self::reader_task(reader, self.tx_events.clone(), cancel.clone()),
+                    Self::writer_task(writer, self.tx_events.clone(), rx_sender, cancel.clone()),
+                )
+            },
+            async {
+                select! {
+                    res = async {
+                        if let Some(tx_shutdown_response) = rx_shutdown.recv().await {
+                            cancel.cancel();
+                            Some(tx_shutdown_response)
+                        } else {
+                            None
+                        }
+                    } => res,
+                    _ = cancel.cancelled() => None
+                }
+            }
+        );
         if let Err(err) = self.tx_events.send(Event::Disconnected).await {
             error!(
                 target: logs::targets::CLIENT,
                 "fail to send event Disconnected; error: {:?}", err
             );
         }
+        let close_err = match writer.reunite(reader) {
+            Ok(mut ws) => ws
+                .close(None)
+                .await
+                .map_err(|e| Error::Closing(e.to_string())),
+            Err(err) => Err(Error::Closing(err.to_string())),
+        };
         self.control.shutdown().await?;
         self.reinit();
+        self.control.get_shutdown_token().cancel();
+        if let Some(tx_shutdown_response) = tx_shutdown_response {
+            if tx_shutdown_response.send(()).is_err() {
+                error!(
+                    target: logs::targets::CLIENT,
+                    "fail to send shutdown response"
+                );
+            }
+        }
         debug!(target: logs::targets::CLIENT, "client is finished");
-        res
+        if let Some(error) = reader_error {
+            Err(error)
+        } else if let Some(error) = writer_error {
+            Err(error)
+        } else {
+            close_err
+        }
     }
 
     fn observer(&mut self) -> Result<Receiver<Event<Error>>, Error> {
